#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#ifndef __FreeBSD__
#include <sys/select.h>
#endif
#include <sys/time.h>
#include <time.h>
#include "dc210.h"

char *cameratypestr[]={"","","Kodak DC50", "Kodak DC120", "Kodak DC210 Zoom"};
char *batterystr[]={"full","low","empty"};
char *compressionstr[]={"none","low","medium","high"};
char *resolutionstr[]={" 640x480", "1152x864"};
char *flashstr[]={"auto","on","off","auto red eye","on w/red eye"};
char *qualstr[]={"","best  ","better","good  "};

int do_command(int argc, char **argv);
int do_list(int argc, char **argv);
int do_status(int argc, char **argv);
int do_get(int argc, char **argv);
int do_tget(int argc, char **argv);
int do_del(int argc, char **argv);
int do_snap(int argc, char **argv);
int do_set(int argc, char **argv);
int do_system(int argc, char **argv);
int do_help(int argc, char **argv);

/* printf to stderr */
void eprintf(char *fmt, ...)
{
    char msg[132];
    va_list args;
    va_start(args, fmt);
    vsprintf(msg, fmt, args);
    fprintf(stderr,msg);
    va_end(args);
}

void usage(int argc, char **argv)
{
	if (argc == 2) {
		fprintf(stderr, "Kodak DC21x Digital Camera utility by timecop "
			"[timecop@japan.co.jp]\n");
		fprintf(stderr, "usage: %s [command [options]]\n", argv[0]);
	}
	do_help(argc - 1, argv + 1);
	return;
}

int main(int argc, char **argv)
{
	enum { CommandLength = 256 };
 	char command[CommandLength], *iargv[10], *c;
 	const char *serdevice;
	int interactive, iargc;

#ifndef __FreeBSD__
 	const char DefaultSerialDevice[] = "/dev/ttyS1";
#else
 	const char DefaultSerialDevice[] = "/dev/cuaa1";
#endif
 	const char PromptString[] = "dcp> ";

	serdevice = DefaultSerialDevice;
	iargc = 1;
	if (argc > 1 && !strncmp(argv[1], "/dev", 4)) {
		serdevice = argv[1];
		iargc = 2;
	}
	interactive = (argc == iargc);

	if (!interactive && !strcmp(argv[iargc], "help")) {
		usage(argc, argv);
		exit(0);
	}

 	serialdev = open_camera(serdevice);
 	camera_init();

	if (!interactive) {
		do_command(argc - iargc, argv + iargc);
		exit(0);
	}

	while ((!isatty(fileno(stdin)) || fprintf(stderr, PromptString)) &&
		fgets(command, CommandLength, stdin)) {
		/* parse command line */
		for (iargc = 0, c = command; *c && iargc < 10; ++iargc) {
			while(*c && isspace(*c)) ++c;
			if (!*c) break;
			iargv[iargc] = c;
			while(*c && !isspace(*c)) ++c;
			if (!*c) break;
			*c++ = '\0';
		}
		do_command(iargc, iargv);
	}

	exit(0);
}

int do_command(int argc, char **argv)
{
	if (argc < 1)
		return 0;
	if (!strcmp(argv[0], "ls") || !strcmp(argv[0], "dir"))
		return do_list(argc, argv);
	else if (!strcmp(argv[0], "stat") || !strcmp(argv[0], "status"))
		return do_status(argc, argv);
	else if (!strcmp(argv[0], "get") || !strcmp(argv[0], "mget"))
		return do_get(argc, argv);
	else if (!strcmp(argv[0], "tget") || !strcmp(argv[0], "mtget"))
		return do_tget(argc, argv);
	else if (!strcmp(argv[0], "del") || !strcmp(argv[0], "mdel"))
		return do_del(argc, argv);
	else if (!strcmp(argv[0], "snap"))
		return do_snap(argc, argv);
	else if (!strcmp(argv[0], "set"))
		return do_set(argc, argv);
	else if (argv[0][0] == '!')
		return do_system(argc, argv);
	else if (!strcmp(argv[0], "help"))
		return do_help(argc, argv);
	else if (!strcmp(argv[0], "quit") || !strcmp(argv[0], "exit"))
		exit(0);
	fprintf(stderr, "%s: unkown command\n", argv[0]);
	return 1;
}

int isinrange(int n, int argc, char **argv)
{
	char c;
	int i, m, mfirst, mlast;

	if (argc > 0 && (!strcmp(argv[0], "all") || !strcmp(argv[0], "*")))
		return 1;

	for (i = 0; i < argc; ++i) {
		if (sscanf(argv[i], "%d-%d", &mfirst, &mlast) == 2) {
			if (n >= mfirst && n <= mlast)
				return 1;
		} else if (sscanf(argv[i], "-%d", &mlast) == 1) { 
			if (n <= mlast)
				return 1;
		} else if (sscanf(argv[i], "%d%c", &mfirst, &c) == 2 && c == '-') { 
			if (n >= mfirst)
				return 1;
		} else if (sscanf(argv[i], "%d", &m) == 1) { 
			if (n == m)
				return 1;
		}
	}
	return 0;
}

int do_list(int argc, char **argv)
{
	enum { BufSize = 256 };
	char buf[BufSize];
	int i;

	get_camera_status();
	for (i = 1; i <= status.num_pictures; ++i) {
		if (argc > 1 && !isinrange(i, argc - 1, argv + 1))
			continue;
		get_picture_info(i);
		strftime(buf, BufSize, "%b %e %H:%M", localtime(&picinfo.elapsedTime));
		fprintf(stderr, "  %3d %s %s %6d %s %-12s\n", i,
			resolutionstr[(int)picinfo.resolution],
			qualstr[(int)picinfo.compression],
			picinfo.fileSize,
			buf,
			picinfo.fileName);
	}
	return 0;
}

int do_status(int argc, char **argv)
{
	enum { BufSize = 256 };
	char buf[BufSize];

	get_camera_status();
	fprintf(stderr,"camera model: %s\n",
		cameratypestr[(int)status.camera_type_id]);
	fprintf(stderr,"camera identification: %-32s\n", status.camera_ident);
	strftime(buf, BufSize, "%+", localtime(&status.time));
	fprintf(stderr,"camera time: %s\n", buf);
	fprintf(stderr,"battery level: %s\n", batterystr[(int)status.battery]);
	fprintf(stderr,"ac power: %sconnected\n", status.acstatus ? "" : "not ");
	fprintf(stderr,"pictures in camera: %d\n", status.num_pictures);
	fprintf(stderr,"flash charged: %s\n", status.flash_charged ? "yes" : "no");
	fprintf(stderr,"flash mode: %s\n", flashstr[(int)status.flash_mode]); 
	fprintf(stderr,"compression mode: %s\n",
			compressionstr[(int)status.compression_mode_id]);
	fprintf(stderr,"total pictures taken: %d\n", status.totalPicturesTaken);
	fprintf(stderr,"total flashes fired: %d\n", status.totalFlashesFired);
	return 0;
}

int do_get(int argc, char **argv)
{
	enum { FileNameLength = 20 };
	char tfname[FileNameLength];
	int multiple = !strcmp(argv[0], "mget");
	int n;

	if (argc < 2) {
		fprintf(stderr, "usage: [m]get n|range [filename]\n");
		return 1;
	}

	get_camera_status();
	if (!multiple) {
		if (sscanf(argv[1], "%d", &n) != 1 || n < 1 ||
			n > status.num_pictures) {
			fprintf(stderr, "invalid picture number: %s\n", argv[1]);
			return 1;
		}
		if (argc < 3) {
			get_picture_info(n);
			strncpy(tfname, picinfo.fileName, 12);
			tfname[12] = '\0';
			fname = tfname;
		} else {
			fname = argv[2];
		}
		get_picture(n);
	} else {
		for (n = 1; n <= status.num_pictures; ++n) {
			if (!isinrange(n, argc - 1, argv + 1))
				continue;
			get_picture_info(n);
			strncpy(tfname, picinfo.fileName, 12);
			tfname[12] = '\0';
			fname = tfname;
			get_picture(n);
		}
	}
	return 0;
}

int do_tget(int argc, char **argv)
{
	enum { FileNameLength = 20 };
	char tfname[FileNameLength], *s;
	int multiple = !strcmp(argv[0], "mtget");
	int n;

	if (argc < 2) {
		fprintf(stderr, "usage: [m]tget n|range [filename]\n");
		return 1;
	}

	get_camera_status();
	if (!multiple) {
		if (sscanf(argv[1], "%d", &n) != 1 || n < 1 ||
			n > status.num_pictures) {
			fprintf(stderr, "invalid picture number: %s\n", argv[1]);
			return 1;
		}
		if (argc < 3) {
			get_picture_info(n);
			strncpy(tfname, picinfo.fileName, 12);
			tfname[12] = '\0';
			if ((s = strrchr(tfname, '.')))
				strcpy(s, ".BMP");
			fname = tfname;
		} else {
			fname = argv[2];
		}
		get_thumbnail(n);
	} else {
		for (n = 1; n <= status.num_pictures; ++n) {
			if (!isinrange(n, argc - 1, argv + 1))
				continue;
			get_picture_info(n);
			strncpy(tfname, picinfo.fileName, 12);
			tfname[12] = '\0';
			if ((s = strrchr(tfname, '.')))
				strcpy(s, ".BMP");
			fname = tfname;
			get_thumbnail(n);
		}
	}
	return 0;
}

int do_del(int argc, char **argv)
{
	int multiple = !strcmp(argv[0], "mdel");
	int n, ndel;

	if (argc < 2) {
		fprintf(stderr, "usage: [m]del n|range\n");
		return 1;
	}

	get_camera_status();
	if (!multiple) {
		if (sscanf(argv[1], "%d", &n) != 1 || n < 1 ||
			n > status.num_pictures) {
			fprintf(stderr, "invalid picture number: %s\n", argv[1]);
			return 1;
		}
		pic_ops(DELE_PIC, n);
		if (!quiet)
			fprintf(stderr, "Deleted image %d from camera.\n", n);
	} else {
		/* after each deletion, the picture numbers decrease by one! */
		ndel = 0;
		for (n = 1; n <= status.num_pictures; ++n) {
			if (!isinrange(n, argc - 1, argv + 1))
				continue;
			pic_ops(DELE_PIC, n - ndel);
			if (!quiet)
				fprintf(stderr, "Deleted image %d from camera.\n", n);
			++ndel;
		}
	}
	return 0;
}

int do_snap(int argc, char **argv)
{
	if (argc == 1) {
		pic_ops(TAKE_PIC, 0);
		return 0;
	}
	pic_ops(TAKE_PIC, 0);
	get_camera_status();
	fname = argv[1];
	get_picture(status.num_pictures);
	pic_ops(DELE_PIC, status.num_pictures);
	return 0;
}

int do_set(int argc, char **argv)
{
	char s[3], packet[58], *p, *q, *packetend;
	int i, zoom;

	if (argc < 2) {
		fprintf(stderr, "usage: set variable [value]\n");
		return 1;
	}

	if (!strcmp(argv[1], "verbose")) {
		quiet = 0;
	} else if (!strcmp(argv[1], "noverbose") || !strcmp(argv[1], "quiet")) {
		quiet = 1;
	} else if (!strcmp(argv[1], "id")) {
		p = packet;
		packetend = packet + 58;
		for (i = 2; i < argc; ++i) {
			q = argv[i];
			while (p < packetend && (*p++ = *q++)) ;
			if (i < argc - 1 && p < packetend) *(p - 1) = ' ';
		}
		while (p < packetend) *p++ = '\0';
		send_command(DC_SET_CAMERA_ID,0x00,0x00,0x00,0x00);
  		write_packet(PKT_CTRL_EOF, packet, 58);
		command_complete();
	} else if (!strcmp(argv[1], "flash")) {
		if (!strcmp(argv[2], "auto")) {
			send_command(DC_SET_FLASH,0x00,0x00,0x00,0x00);
			command_complete();
		} else if (!strcmp(argv[2], "on")) {
			send_command(DC_SET_FLASH,0x01,0x00,0x00,0x00);
			command_complete();
		} else if (!strcmp(argv[2], "off")) {
			send_command(DC_SET_FLASH,0x02,0x00,0x00,0x00);
			command_complete();
		} else if (!strcmp(argv[2], "autoredeye")) {
			send_command(DC_SET_FLASH,0x03,0x00,0x00,0x00);
			command_complete();
		} else if (!strcmp(argv[2], "redeye")) {
			send_command(DC_SET_FLASH,0x04,0x00,0x00,0x00);
			command_complete();
		}
	} else if (!strcmp(argv[1], "quality")) {
		if (!strcmp(argv[2], "best")) {
			send_command(DC_SET_QUALITY,0x01,0x00,0x00,0x00);
			command_complete();
		} else if (!strcmp(argv[2], "better")) {
			send_command(DC_SET_QUALITY,0x02,0x00,0x00,0x00);
			command_complete();
		} else if (!strcmp(argv[2], "good")) {
			send_command(DC_SET_QUALITY,0x03,0x00,0x00,0x00);
			command_complete();
		}
	} else if (!strcmp(argv[1], "resolution")) {
		if (!strcmp(argv[2], "low")) {
			send_command(DC_SET_RESOLUTION,0x00,0x00,0x00,0x00);
			command_complete();
		} else if (!strcmp(argv[2], "high")) {
			send_command(DC_SET_RESOLUTION,0x01,0x00,0x00,0x00);
			command_complete();
		}
	} else if (!strcmp(argv[1], "zoom")) {
		if (sscanf(argv[2], "%d%2s", &zoom, s) == 2 && !strcmp(s, "mm")) {
			switch (zoom) {
				case 58:
					send_command(DC_SET_ZOOM,0x00,0x00,0x00,0x00);
					command_complete();
					break;
				case 51:
					send_command(DC_SET_ZOOM,0x01,0x00,0x00,0x00);
					command_complete();
					break;
				case 41:
					send_command(DC_SET_ZOOM,0x02,0x00,0x00,0x00);
					command_complete();
					break;
				case 34:
					send_command(DC_SET_ZOOM,0x03,0x00,0x00,0x00);
					command_complete();
					break;
				case 29:
					send_command(DC_SET_ZOOM,0x04,0x00,0x00,0x00);
					command_complete();
					break;
			}
		} else if (!strcmp(argv[2], "macro")) {
			send_command(DC_SET_ZOOM,0x05,0x00,0x00,0x00);
			command_complete();
		}
	}
	return 0;
}

int do_system(int argc, char **argv)
{
	int i;
	char command[256], *c, *d, *commandend;

	c = command;
	commandend = command + 256;
	for (i = (strcmp(argv[0], "!") == 0); i < argc; ++i) {
		d = i ? argv[i] : argv[i] + 1;
		while (c < commandend && (*c++ = *d++)) ;
		if (i < argc - 1 && c < commandend) *(c - 1) = ' ';
	}
	*(commandend - 1) = '\0';
	return system(command);
}

int do_help(int argc, char **argv)
{
	int i;

	const char *CommandList[][2] = {
		{ "!", "escape to the shell" },
		{ "del", "delete picture - del <np>" },
		{ "dir", "list contents of camera directory - dir [range [range]..]" },
		{ "exit", "terminate dcp session and exit" },
		{ "get", "download picture - get <np> [filename]" },
		{ "help", "display this help" },
		{ "ls", "list contents of camera directory - ls [range [range]..]" },
		{ "mdel", "delete multiple pictures - mdel [range [range]..]" },
		{ "mget", "download multiple pictures - mget [range [range]..]" },
		{ "mtget", "download multiple thumbnails - mget [range [range]..]" },
		{ "quit", "terminate dcp session and exit" },
		{ "set", "set a variable - set variable [value]\n"
			"\tvariables are:\n"
			"\t    verbose\n"
			"\t    noverbose|quiet\n"
			"\t    id <camera id>\n"
			"\t    flash auto|on|off|autoredye|redeye\n"
			"\t    quality good|better|best\n"
			"\t    resolution low|high\n"
			"\t    zoom 58mm|51mm|41mm|34mm|29mm|macro" },
		{ "snap", "take a picture - snap [filename]" },
		{ "status", "show current camera status" },
		{ "tget", "download thumbnail - tget <np> [filename]" }
	};

	if (argc <= 1)
		fprintf(stderr, "Commands are:\n");

	for (i = 0; i < sizeof(CommandList) / sizeof(CommandList[0]); ++i) {
		if (argc <= 1 || !strcmp(argv[1], CommandList[i][0]))
			fprintf(stderr, "%s\t%s\n", CommandList[i][0], CommandList[i][1]);
	}

	return 0;
}
